<?php
public function getProjectRoles()
{
    $projectRoles  = '';
    $approvalFlows = $this->dao->select('id,version')->from(TABLE_APPROVALFLOW)->where('deleted')->eq('0')->fetchAll();
    $specs         = $this->dao->select('flow,version,nodes')->from(TABLE_APPROVALFLOWSPEC)->fetchAll('version');

    foreach($approvalFlows as $approvalFlow)
    {
        $nodes = json_decode($specs[$approvalFlow->version]->nodes);

        foreach($nodes as $node)
        {
            if(isset($node->reviewers))
            {
                $projectRole = array_column($node->reviewers, 'projectRoles');
                foreach($projectRole as $role) $projectRoles .= join(',', $role) . ',';
            }
            if(isset($node->ccs))
            {
                $projectRole = array_column($node->ccs, 'projectRoles');
                foreach($projectRole as $role) $projectRoles .= join(',', $role) . ',';
            }
        }
    }
    $projectRoles = trim($projectRoles, ',');

    return array_unique(explode(',', $projectRoles));
}

/**
 * Get fields of an node.
 *
 * @param  string $module
 * @param  object $action
 * @param  bool   $getRealOptions
 * @param  array  $datas
 * @access public
 * @return array
 */
public function getNodeFields($module, $action, $flowID, $getRealOptions = true, $datas = array())
{
    $actionFields = $this->dao->select("t1.*, t2.id, t2.name, t2.control, t2.expression, t2.options, t2.type, t2.length, t2.rules, t2.placeholder, t2.canExport, t2.canSearch, t2.isValue, t2.desc, t2.buildin, t2.default")
        ->from(TABLE_WORKFLOWNODELAYOUT)->alias('t1')
        ->leftJoin(TABLE_WORKFLOWFIELD)->alias('t2')->on('t1.module=t2.module AND t1.field=t2.field')
        ->where('t1.module')->eq($module)
        ->andWhere('t1.action')->eq($action)
        ->andWhere('t1.flow')->eq($flowID)
        ->beginIF(!empty($this->config->vision))->andWhere('t1.vision')->eq($this->config->vision)->fi()
        ->orderBy('t1.order, t2.order, t2.id')
        ->fetchAll('field');

    $datas = $this->loadModel('workflowaction')->mergeDefaultDatas($actionFields, $datas);
    $flow  = $this->loadModel('workflow', 'flow')->getByModule($module);
    if($flow and $flow->type == 'flow') $action = $this->loadModel('workflowaction')->getByModuleAndAction($module, $action);

    $subTables = $this->workflow->getPairs($module, $type = 'table');
    $fieldList = $this->loadModel('workflowfield', 'flow')->getList($module);
    $fieldList = $this->loadModel('workflowaction')->processFields($fieldList, $getRealOptions, $datas);

    foreach($actionFields as $key => $field)
    {
        $field = $this->loadModel('workflowfield')->processFieldOptions($field);

        if(!isset($fieldList[$field->field])) continue;

        $field->options = $fieldList[$field->field]->options;
    }

    if(empty($actionFields))
    {
        foreach($fieldList as $field)
        {
            if(!$field) continue;
            if(zget($flow, 'type') == 'table' && $field->field == 'mailto') continue;

            $show         = 0;
            $mobileShow   = 0;
            $defaultValue = '';
            $layoutRules  = '';

            if(in_array($field->field, array('reviewResult', 'reviewOpinion')))
            {
                $show       = 1;
                $mobileShow = 1;
                if($field->field == 'reviewResult') $defaultValue = 'pass';
                if($field->field == 'reviewOpinion') $layoutRules = '1';
            }

            $fields[$field->field] = $this->loadModel('workflowaction')->getNewField();
            $fields[$field->field]->id           = $field->id;
            $fields[$field->field]->module       = $module;
            $fields[$field->field]->field        = $field->field;
            $fields[$field->field]->type         = $field->type;
            $fields[$field->field]->length       = $field->length;
            $fields[$field->field]->name         = $field->name;
            $fields[$field->field]->control      = $field->control;
            $fields[$field->field]->expression   = $field->expression;
            $fields[$field->field]->buildin      = $field->buildin;
            $fields[$field->field]->canExport    = $field->canExport;
            $fields[$field->field]->canSearch    = $field->canSearch;
            $fields[$field->field]->isValue      = $field->isValue;
            $fields[$field->field]->options      = $field->options;
            $fields[$field->field]->default      = $field->default;
            $fields[$field->field]->show         = $show;
            $fields[$field->field]->mobileShow   = $mobileShow;
            $fields[$field->field]->defaultValue = $defaultValue;
            $fields[$field->field]->layoutRules  = $layoutRules;
        }

        if(zget($action, 'method') == 'browse' && zget($action, 'extensionType') == 'override')
        {
            $fields['actions'] = $this->loadModel('workflowaction')->getNewField();
            $fields['actions']->id     = 'actions';
            $fields['actions']->module = $module;
            $fields['actions']->field  = 'actions';
            $fields['actions']->name   = $this->lang->actions;
            $fields['actions']->show   = '1';
            $fields['actions']->width  = 'auto';
        }

        if(zget($flow, 'type') == 'flow' && zget($action, 'method') != 'browse' && zget($action, 'type') != 'batch')
        {
            foreach($subTables as $subModule => $subName)
            {
                /* 为了把明细表代号和主表字段区分开，增加前缀。In order to distinguish the schedule code from the main table field, add a prefix. */
                $subModule = 'sub_' . $subModule;

                $fields[$subModule] = $this->loadModel('workflowaction')->getNewField();
                $fields[$subModule]->id     = $subModule;
                $fields[$subModule]->module = $module;
                $fields[$subModule]->field  = $subModule;
                $fields[$subModule]->name   = $subName;
            }
        }
    }
    else
    {
        foreach($actionFields as $key => $field)
        {
            if(!$field) continue;

            if($field->layoutRules)
            {
                $rules = "{$field->rules},{$field->layoutRules}";
                $rules = explode(',', $rules);

                $field->rules = implode(',', array_unique($rules));
            }

            $fields[$key] = $this->loadModel('workflowaction')->getNewField();
            $fields[$key]->id           = ($key == 'actions') ? $key : $field->id;
            $fields[$key]->module       = $module;
            $fields[$key]->field        = $field->field;
            $fields[$key]->type         = $field->type;
            $fields[$key]->length       = $field->length;
            $fields[$key]->name         = zget($subTables, str_replace('sub_', '', $key), $field->name);    // 把前缀去掉来获得明细表名。
            $fields[$key]->control      = $field->control;
            $fields[$key]->expression   = $field->expression;
            $fields[$key]->buildin      = $field->buildin;
            $fields[$key]->canExport    = $field->canExport;
            $fields[$key]->canSearch    = $field->canSearch;
            $fields[$key]->isValue      = $field->isValue;
            $fields[$key]->show         = '1';
            $fields[$key]->width        = $field->width ? $field->width : 'auto';
            $fields[$key]->position     = $field->position;
            $fields[$key]->readonly     = $field->readonly;
            $fields[$key]->mobileShow   = $field->mobileShow;
            $fields[$key]->summary      = $field->summary;
            $fields[$key]->defaultValue = $field->defaultValue;
            $fields[$key]->rules        = $field->rules;
            $fields[$key]->sql          = $field->sql;
            $fields[$key]->layoutRules  = $field->layoutRules;
            $fields[$key]->options      = $field->options;
            $fields[$key]->default      = $field->default;

            if($key == 'actions') $fields[$key]->name = $this->lang->actions;
            if($key == 'actions' && zget($action, 'method') == 'browse') $fields[$key]->show = 1;
        }

        foreach($fieldList as $id => $field)
        {
            if(!$field or isset($actionFields[$field->field])) continue;
            if(zget($flow, 'type') == 'table' && $field->field == 'mailto') continue;

            $fields[$field->field] = $this->loadModel('workflowaction')->getNewField();
            $fields[$field->field]->id         = $field->id;
            $fields[$field->field]->module     = $module;
            $fields[$field->field]->field      = $field->field;
            $fields[$field->field]->type       = $field->type;
            $fields[$field->field]->length     = $field->length;
            $fields[$field->field]->name       = $field->name;
            $fields[$field->field]->control    = $field->control;
            $fields[$field->field]->expression = $field->expression;
            $fields[$field->field]->buildin    = $field->buildin;
            $fields[$field->field]->canExport  = $field->canExport;
            $fields[$field->field]->canSearch  = $field->canSearch;
            $fields[$field->field]->isValue    = $field->isValue;
            $fields[$field->field]->options    = $field->options;
            $fields[$field->field]->default    = $field->default;
        }

        if(zget($action, 'method') == 'browse' && zget($action, 'extensionType') == 'override' && !isset($fields['actions']))
        {
            $fields['actions'] = $this->loadModel('workflowaction')->getNewField();
            $fields['actions']->id      = 'actions';
            $fields['actions']->module  = $module;
            $fields['actions']->field   = 'actions';
            $fields['actions']->name    = $this->lang->actions;
            $fields['actions']->show    = '1';
            $fields['actions']->width   = 'auto';
            $fields['actions']->default = '';
        }

        if(zget($flow, 'type') == 'flow' && zget($action, 'method') != 'browse' && zget($action, 'type') != 'batch')
        {
            foreach($subTables as $subModule => $subName)
            {
                if(!$subModule) continue;

                /* 为了把明细表代号和主表字段区分开，增加前缀。In order to distinguish the schedule code from the main table field, add a prefix. */
                if($subModule) $subModule = 'sub_' . $subModule;
                if(isset($actionFields[$subModule])) continue;

                $fields[$subModule] = $this->loadModel('workflowaction')->getNewField();
                $fields[$subModule]->id      = $subModule;
                $fields[$subModule]->module  = $module;
                $fields[$subModule]->field   = $subModule;
                $fields[$subModule]->name    = $subName;
                $fields[$subModule]->default = '';
            }
        }
    }

    return $fields;
}

/**
 * Get unbound approval flow.
 *
 * @param  string $workflowModule
 * @access public
 * @return array
 */
public function getUnboundApprovalFlow($workflowModule = '')
{
    $bindWorkFlow = $this->config->workflow->buildin->noApproval;

    $approvalFlows = $this->dao->select('module,name')->from(TABLE_WORKFLOW)
        ->where('type')->eq('flow')
        ->beginIF($workflowModule)
        ->andWhere('(approval')->eq('disabled')
        ->orWhere('module')->eq($workflowModule)
        ->markRight('1')
        ->fi()
        ->beginIF(!$workflowModule)->andWhere('approval')->eq('disabled')->fi()
        ->beginIF($bindWorkFlow)->andWhere('module')->notin($bindWorkFlow)->fi()
        ->fetchPairs('module');

    return array('' => '') + $approvalFlows;
}

/**
 * Get work flow module.
 *
 * @param  int $flowID
 * @access public
 * @return string
 */
public function getWorkflowModule($flowID = 0)
{
    $module = $this->dao->select('objectType')->from(TABLE_APPROVALFLOWOBJECT)->where('flow')->eq($flowID)->orderBy('id_desc')->fetch('objectType');

    if($module)
    {
        $workflowApproval = $this->dao->select('approval')->from(TABLE_WORKFLOW)->where('module')->eq($module)->orderBy('id_desc')->fetch('approval');
        $module = $workflowApproval == 'enabled' ? $module : '';
    }

    return $module;
}

/**
 * Get work flow set fields by flow id.
 *
 * @param  int $flowID
 * @access public
 * @return array
 */
public function getSetFieldsByflowID($flowID = 0)
{
    return $this->dao->select('*')->from(TABLE_WORKFLOWNODELAYOUT)->where('flow')->eq($flowID)->fetchAll();
}

/**
 * Get unbound work flow paris.
 *
 * @param  string $approvalFlow
 * @access public
 * @return array
 */
public function getUnboundPairs($approvalFlow = '')
{
    $bindApprovalFlows = $this->dao->select('t2.flow')->from(TABLE_WORKFLOW)->alias('t1')
        ->leftJoin(TABLE_APPROVALFLOWOBJECT)->alias('t2')->on('t1.module = t2.objectType')
        ->where('t1.approval')->eq('enabled')
        ->fetchPairs();

    if($approvalFlow and isset($bindApprovalFlows[$approvalFlow])) unset($bindApprovalFlows[$approvalFlow]);

    $approvalFlows = $this->dao->select('id,name')->from(TABLE_APPROVALFLOW)
        ->where('deleted')->eq('0')
        ->andWhere('type')->eq('workflow')
        ->beginIF($bindApprovalFlows)->andWhere('id')->notin($bindApprovalFlows)->fi()
        ->fetchPairs();

    return array('' => '') + $approvalFlows;
}
